__version__ = "0.0.5"

from motorodm.documents import *
from motorodm.fields import *

__all__ = [
    "Document",
    "EmbeddedDocument",

    "BooleanField",
    "DateTimeField",
    "DecimalField",
    "EmbeddedDocumentField",
    "FloatField",
    "IntField",
    "JsonField",
    "ListField",
    "ObjectIdField",
    "ReferenceField",
    "StringField"
]



from .field import Field


class IntField(Field):

    def __init__(self, min_value=None, max_value=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.min_value = min_value
        self.max_value = max_value

    def validate(self, value):
        if value is not None:
            if not isinstance(value, int):
                return False

            if self.min_value is not None and value < self.min_value:
                return False

            if self.max_value is not None and value > self.max_value:
                return False

        return super().validate(value)

    def is_empty(self, value):
        return value is None



from ..fields.field import Field
from ..fields import ObjectIdField
from ..query_sets.query_set import QuerySet


class MetaEmbeddedDocument(type):

    def __new__(cls, name, bases, dct):

        dct['_fields'] = {}
        dct['_db_name_map'] = {}
        dct['_indices'] = []

        for base in bases:
            for field_name, field in filter(lambda x: isinstance(x[1], Field), base.__dict__.items()):
                cls.add_field(dct, field_name, field)

        for field_name, field in filter(lambda x: isinstance(x[1], Field), dct.items()):
            field.name = field_name
            cls.add_field(dct, field_name, field)

        dct['_values'] = {}
        dct['_dirty_fields'] = set()

        return super().__new__(cls, name, bases, dct)

    @classmethod
    def add_field(cls, dct, field_name, field):
        if field_name in dct['_fields']:
            raise KeyError(f"Field '{field_name}' already exists")
        if not field.db_name:
            field.db_name = field_name
        if field.db_name in dct['_db_name_map']:
            raise KeyError(f"Field '{field_name}' already exists")
        field.name = field_name
        dct['_fields'][field_name] = field
        dct['_db_name_map'][field.db_name] = field_name
        if field.unique:
            dct['_indices'].append(field_name)


class MetaDocument(MetaEmbeddedDocument):

    def __new__(cls, name, bases, dct):

        if '_root' in dct and dct['_root']:
            return super().__new__(cls, name, bases, dct)

        if dct.get('__collection__', None) is None:
            dct['__collection__'] = name

        klass = super().__new__(cls, name, bases, dct)

        if '_id' not in klass._db_name_map:
            if 'id' in klass.__dict__:
                raise Exception('Unable to set id field - already exists')
            field = ObjectIdField(name='id', db_name='_id')
            klass.id = field
            klass._fields[field.name] = field
            klass._db_name_map[field.db_name] = field.name

        klass.qs = QuerySet()

        return klass



class DocumentIterator:

    def __init__(self, cursor, document_class, db):
        self.cursor = cursor
        self.document_class = document_class
        self.db = db


    def __aiter__(self):
        return self


    async def __anext__(self):
        if await self.cursor.fetch_next:
            data = self.cursor.next_object()
            return await self.document_class.from_mongo(data, self.resolve)
        else:
            raise StopAsyncIteration


    async def distinct(self):
        return await self.cursor.distinct()


    def limit(self, n):
        self.cursor.limit(n)


    def skip(self, n):
        self.cursor.skip(n)


    def sort(self, key_or_list, direction=None):
        self.cursor.sort(key_or_list, direction)


    async def to_list(self, length):
        return [await self.document_class.from_mongo(item, self.resolve) for item in await self.cursor.to_list(length)]


    async def resolve(self, document_class, value):
        value = await document_class.qs(self.db).get(value)
        return value


    @property
    def alive(self):
        return self.cursor.alive



from .boolean_field import BooleanField
from .datetime_field import DateTimeField
from .decimal_field import DecimalField
from .embedded_document_field import EmbeddedDocumentField
from .float_field import FloatField
from .int_field import IntField
from .json_field import JsonField
from .list_field import ListField
from .object_id_field import ObjectIdField
from .reference_field import ReferenceField
from .string_field import StringField

__all__ = [
    "BooleanField",
    "DateTimeField",
    "DecimalField",
    "EmbeddedDocumentField",
    "FloatField",
    "IntField",
    "JsonField",
    "ListField",
    "ObjectIdField",
    "ReferenceField",
    "StringField"
]



from .field import Field


class ListField(Field):

    def __init__(self, item_field=None, *args, **kwargs):

        if not isinstance(item_field, Field):
            raise ValueError("The item_field for must be a Field")

        super().__init__(*args, **kwargs)

        self.item_field = item_field

    def validate(self, value):
        if value is not None:
            if not isinstance(value, list):
                return False

            for item in value:
                if not self.item_field.validate(item):
                    return False

        return super().validate(value)

    def is_empty(self, value):
        return value is None or len(value) == 0

    def to_mongo(self, value):
        return None if value is None else [self.item_field.to_mongo(item) for item in value]

    async def from_mongo(self, value, resolver=None):
        if value is None:
            return []
        return [await self.item_field.from_mongo(item, resolver) for item in value]



from .field import Field


class ReferenceField(Field):

    def __init__(self, reference_document_type=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reference_document_type = reference_document_type


    def validate(self, value):
        if value is not None:
            if not isinstance(value, self.reference_document_type):
                return False

        return super().validate(value)


    def to_mongo(self, value):
        return value._identity


    async def from_mongo(self, value, resolver=None):
        value = await resolver(self.reference_document_type, value)
        return value



import decimal
from .field import Field


class DecimalField(Field):

    def __init__(self, min_value=None, max_value=None, precision=2, rounding=decimal.ROUND_HALF_UP, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.min_value = min_value
        if self.min_value is not None:
            self.min_value = decimal.Decimal(min_value)

        self.max_value = max_value
        if self.max_value is not None:
            self.max_value = decimal.Decimal(max_value)

        self.precision = decimal.Decimal(".%s" % ("0" * precision))
        self.rounding = rounding

    def validate(self, value):
        if value is not None:
            if not isinstance(value, float):
                return False

            if self.min_value is not None and value < self.min_value:
                return False

            if self.max_value is not None and value > self.max_value:
                return False

        return super().validate(value)

    def is_empty(self, value):
        return value is None

    def to_mongo(self, value):
        value = decimal.Decimal(value)
        return str(value.quantize(self.precision, rounding=self.rounding))

    async def from_mongo(self, value, resolver=None):
        value = decimal.Decimal(value)
        return value.quantize(self.precision, rounding=self.rounding)



from .field import Field


class EmbeddedDocumentField(Field):

    def __init__(self, embedded_document_type=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.embedded_document_type = embedded_document_type

    def validate(self, value):
        if value is not None:
            if not isinstance(value, self.embedded_document_type):
                return False

        return super().validate(value)

    def to_mongo(self, value):
        return self.embedded_document_type.to_mongo(value)

    async def from_mongo(self, value, resolver=None):
        return await self.embedded_document_type.from_mongo(value, resolver)
